//============================================================================
// Name        : FIR_SSE.cpp
// Author      : Daniel
// Version     :
// Copyright   : 
//============================================================================

// c++ standart zeug
#include <iostream>
#include <stdint.h>
#include <string.h>
#include <malloc.h>

// SSE
#include <xmmintrin.h>

// ***********************************************************************************************************

#define _ALG __attribute__((aligned (16)))
typedef float sample_t[2] _ALG;

// ***********************************************************************************************************

inline void conv_step_complex(sample_t sample, const sample_t *pulse, sample_t *mem,  const uint32_t length);
inline void gen_rect_complex(sample_t *mem, uint32_t from, uint32_t to);

// ***********************************************************************************************************
int main() {

	const uint32_t size_sig = 32;			// 32 complexe float samples => 16 SSE werte! muss ohne rest durch zwei teilbar sein!
	const uint32_t size_mem = 8;			// 8 complexe float pulsform koeffizienten => 4 SSE werte! muss ohne rest durch zwei teilbar sein!
	const uint32_t size_pulse = size_mem;

	// signal samples.
	sample_t *samples = (sample_t*) memalign(16, size_sig * sizeof(sample_t));
	memset(samples, 0, size_sig * sizeof(sample_t));

	// pulsform zum filtern (koeffizienten des filters).
	sample_t *pulse = (sample_t*) memalign(16, size_pulse * sizeof(sample_t));
	memset(pulse, 0, size_pulse * sizeof(sample_t));

	// speicher für die filterung (speicherzellen des filters).
	sample_t *mem = (sample_t*) memalign(16, size_mem * sizeof(sample_t));
	memset(mem, 0, size_mem * sizeof(sample_t));



	// pulsform generieren.
	gen_rect_complex(pulse, 0, 5);

	fprintf(stderr,"\nPulsform (nur Realteil): ");
	for(uint32_t j = 0; j < size_pulse; j++){
		fprintf(stderr,"%2.0f ",pulse[j][0]);
	}


	// signal generieren.
	gen_rect_complex(samples, 9, 5);

	fprintf(stderr,"\nSignal (nur Realteil): ");
	for(uint32_t j = 0; j < size_sig; j++){
		fprintf(stderr,"%2.0f ",samples[j][0]);
	}



	// signal mit pulsform filtern.
	for(uint32_t step = 0; step < size_sig; step++){

		sample_t sample _ALG;
		memcpy(sample,  samples + step, sizeof(sample_t));

		conv_step_complex(sample, pulse, mem, size_mem);

		memcpy(samples + step, sample, sizeof(sample_t));
	}

	fprintf(stderr,"\n\nErgebniss (nur Realteil): ");
	for(uint32_t j = 0; j < size_sig; j++){
		fprintf(stderr,"%2.0f ",samples[j][0]);
	}
	fprintf(stderr,"\n\n");

	free(samples);
	free(pulse);
	free(mem);

	return 0;
}

// ***********************************************************************************************************
// nicht optimal, weil nicht als ringbuffer (bei langen filtern ist memmove keine gute idee)...
// ***********************************************************************************************************
inline void conv_step_complex(sample_t sample, const sample_t *pulse, sample_t *mem, const uint32_t length) {

	// speicher für ergebniss aus sse beschleunigter berechnung.
	float res[4] _ALG;

	// 0te speicherstelle beschreiben (am ende der funktion werden alle um 1 sample_t nach rechts gerückt).
	memcpy(mem, sample, sizeof(sample_t));

	// 2 sample_t entsprechen einem sse datentyp (byte anzahl).
	uint32_t step = length;

	// accumulator.
	__m128 sum = _mm_setzero_ps();

	// sse 'beschleunigte' schleife.
	do{
		step -= 2;
		sum = _mm_add_ps(sum, _mm_mul_ps( _mm_load_ps((float*)(mem + step)) , _mm_load_ps((float*)(pulse + step)) ));
	}while(step);

	_mm_store_ps(res, sum);

	// ergebnis auf sample schreiben
	sample[0] = res[0] + res[2]; // realteil.
	sample[1] = res[1] + res[3]; // imaginärteil.

	// alle werte in speicher um 1 sample_t nach rechts schieben. so kann beim nächsten durchgang
	// einfach das neue sample_t auf stelle 0 gesetzt werden.
	memmove(mem + 1, mem, (length - 1)* sizeof(sample_t)); //dest, src, n-bytes
}

// ***********************************************************************************************************
inline void gen_rect_complex(sample_t *mem, uint32_t from, uint32_t len){
	for(uint32_t i = from; i < from + len; i++){
		mem[i][0] = 1.0f;
		mem[i][1] = 1.0f;
	}
}





























